/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "chan_list.hh"
#include "ParseLine.hh"
#include <stdexcept>
#include <iostream>
#include <sstream>

//======================================  Make a fast swap template
namespace std {
    template<>
    inline void
    swap(sends::chan_entry::source_entry& a, 
	 sends::chan_entry::source_entry& b) {
	a.swap(b);
    }

    template<>
    inline void
    swap(sends::chan_entry& a, sends::chan_entry& b) {
	a.swap(b);
    }
}

using namespace std;
using namespace sends;

//======================================  String size for memory estimates
index_type
string_size(const std::string& str) {
    return (str.size() + 7) & (-8);
}

//======================================  Channel entry constructor
//  static const gps_type inf_gps=3999999999;

//======================================  Source entry null constructor
chan_entry::source_entry::source_entry(void)
    : _start(0), _end(0)
{}

//======================================  Source entry data constructor
chan_entry::source_entry::source_entry(const std::string& name, 
				       gps_type start,  gps_type end)
    : _name(name), _start(start), _end(end)
{}

//======================================  Source entry destructor
chan_entry::source_entry::~source_entry(void) {
}

//======================================  Merge source entries
void 
chan_entry::source_entry::merge(const source_entry& se) {
    if (se._start < _start) _start = se._start;
    if (se._end   > _end)   _end   = se._end;
}

//======================================  Swap source entries
void
chan_entry::source_entry::swap(source_entry& b) {
    _name.swap(b._name);
    std::swap<gps_type> (_start, b._start);
    std::swap<gps_type> (_end,   b._end);
}

//======================================  Channel entry constructor
chan_entry::chan_entry(const string& name, chan_type type, long access, 
		       data_type dtype, double rate, const string& source,
		       gps_type start, gps_type end) 
    : mName(name), mType(type), mAccess(access), mDType(dtype), mRate(rate)
{
    if (!source.empty()) addSource(source, start, end);
}

//======================================  Test if channel is available at time.
//
//  This is a stub function - univeral availablility is cool.
bool 
chan_entry::available(gps_type t) const {
    return true;
}

//======================================  Add a source entry
void 
chan_entry::addSource(const std::string& ftype, gps_type start, gps_type end) {
    source_entry newent(ftype, start, end);
    source_iter itr = mSource.lower_bound(newent);
    if (itr != mSource.begin() && (itr-1)->overlap(newent)) {
	source_iter jtr = --itr;
	jtr->merge(newent);
	if (jtr->overlap(*itr)) {
	    jtr->merge(*itr);
	    mSource.erase(itr);
	}
    } else if (itr != mSource.end() && itr->overlap(newent)) {
	itr->merge(newent);
    } else {
	//------------------------------  Save recalculating the lower_bound
	static_cast<source_vect::base_type&>(mSource).insert(itr, newent);
    }
}

//======================================  Search source list value
bool 
chan_entry::is_source(const std::string& src, gps_type t) const {
    source_count N = mSource.size();
    for (source_count i=0; i<N; ++i) {
	if (mSource[i].name() == src && mSource[i].test_gps(t)) return true;
    }
    return false;
}

//======================================  Merge source entries
void 
chan_entry::merge(const chan_entry& ce) {
    mSource.merge(ce.mSource);
}

//======================================  Memory usage
index_type 
chan_entry::mem_use(void) const {
    index_type nBytes = sizeof(chan_entry) + string_size(name());
    index_type N = mSource.size();
    for (index_type i=0; i<N; ++i) {
	nBytes += sizeof(source_entry) + string_size(mSource[i].name());
    }
    return nBytes;
}

//======================================  Swap two channel entries
void
chan_entry::swap(chan_entry& ce) {
    std::swap<chan_type>(mType,   ce.mType);
    std::swap<long>     (mAccess, ce.mAccess);
    std::swap<data_type>(mDType,  ce.mDType);
    std::swap<double>   (mRate,   ce.mRate);
    mName.swap(ce.mName);
    mSource.swap(ce.mSource);
}

//======================================  Set the access level.
void
chan_entry::set_access(long lvl) {
    mAccess = lvl;
}

//======================================  Write a channel description
std::ostream& 
chan_entry::write(ostream& out) const {
    out << mName << " " << mType.str() << " " << mAccess << " " 
	<< mDType.str() << " " << mRate;
    index_type N = mSource.size();
    for (index_type i=0; i<N; ++i) out << " " << mSource[i].name();
    out << endl;
    return out;
}



//======================================  Merge source lists
std::ostream& 
sends::operator<<(std::ostream& out, const chan_entry& ent) {
    return out << ent.name() << "%" 
	       << ent.type().str();
}

//======================================  Channel entry constructor
chan_list::chan_list(unsigned long ndef) 
    : list_available(false)
{
    mChanList.reserve(ndef);
}

//======================================  Add a channel to the list.
chan_index
chan_list::add_channel(const chan_entry& ent) {
    chan_index inx = mChanList.findex(ent);
    if (inx == mChanList.size()) mChanList.insert(ent);
    else	                 mChanList[inx].merge(ent);
    return inx;
}

//======================================  Remove all entries from the list.
void 
chan_list::clear(void) {
    list_available = false;
    mChanList.clear();
}

//======================================  Print all entries from the list.
std::ostream& 
chan_list::dump(std::ostream& out) const {
    out << "Channel list: " << endl;
    out << "Name  type  Source  access" << endl;
    chan_index N=size();
    for (chan_index i=0; i<N; ++i) {
	out << mChanList[i].name() << "  " << mChanList[i].type().str()
	    << "  " << mChanList[i].getAccess();
	int M = mChanList[i].getNSource();
	for (int j=0; j<M; ++j) out << "  " <<  mChanList[i].getSource(j);
	out << endl;
    }
    return out;
}

//======================================  Allow access to the channel list
void 
chan_list::enable(void) {
    list_available = true;
}

//======================================  Find the specified entry
chan_index 
chan_list::find_best(const string& name, chan_type t,
		     double rate, gps_type gps) const {
    chan_index lb = lower_bound(name, t);
    chan_index ub = upper_bound(name, t);
    for (chan_index inx=lb; inx<ub; ++inx) {
	if (rate != 0 && rate > mChanList[inx].rate()) continue;
	if (gps  != 0) {
	    if (mChanList[inx].type() == chan_type::kOnline) continue;
	    if (!mChanList[inx].available(gps)) continue;
	}
	return inx;
    }
    throw runtime_error(string("Unknown or unavailable channel: ") + name);
}

//======================================  Find the specified entry
chan_index 
chan_list::lower_bound(const string& name, chan_type t) const {
    chan_entry cent(name, t);
    return mChanList.lower_bound(cent) - mChanList.begin();
}

//======================================  Find the specified entry
chan_index 
chan_list::upper_bound(const string& name, chan_type t) const {
    chan_entry cent(name, t);
    return mChanList.upper_bound(cent) - mChanList.begin();
}

//======================================  Merge a channel
void
chan_list::merge(const chan_vect::base_type& vec) {
    index_type N = vec.size();
    if (!N) return;
    bool* newch = new bool[N];
    index_type nnew = 0;

    index_type end = size();
    for (index_type i=0; i<N; ++i) {
	index_type inx = find(vec[i]);
	if (inx == end) {
	    newch[i] = true;
	    nnew++;
	} else {
	    newch[i] = false;
	    mChanList[inx].merge(vec[i]);
	}
    }

    //----------------------------------  Merge new entries into the list.
    if (nnew) {
	mChanList.reserve(size() + nnew);
	for (index_type i=0; i<N; i++) {
	    if (newch[i]) mChanList.push_back(vec[i]);
	}
	mChanList.sort();
    }
    delete[] newch;
}

//======================================  Memory usage
index_type 
chan_list::mem_use(void) const {
    index_type nBytes = sizeof(chan_list);
    index_type N = mChanList.size();
    for (index_type i=0; i<N; ++i) {
	nBytes += mChanList[i].mem_use();
    }
    return nBytes;
}

//======================================  Read channels from a file
void 
chan_list::read(const std::string& file, gps_type start, gps_type end) {
    index_type  vect_size = 1024;
    chan_vect::base_type vect;
    vect.reserve(vect_size);

    ParseLine pl(file.c_str());
    if (!pl.isOpen()) {
	cout << "*** Error: Missing channel list file: " << file <<endl;
	return;
    }

    while (pl.getLine() >= 0) {
	int nArg = pl.getCount();
	if (!nArg) continue;
	if (nArg < 2) {
	    ostringstream msg;
	    msg << "chan_list::read: Insufficient arguments in "
	        << file << ": " << pl.getLineNumber();
	    throw runtime_error(msg.str());
	}
	string name = pl[0];
	chan_type type(pl[1]);

	//------------------------------  Get access level
	long access = 0;
	if (nArg > 2) access = pl.getInt(2);

	//------------------------------  Get data type
	data_type dtype(data_type::kReal4);
	if (nArg > 3 && *pl[3] != '-') {
	    dtype.set(pl[3]);
	}

	//------------------------------  Get data rate
	double rate = 0.0;
	if (nArg > 4 && *pl[4] != '-') {
	    rate = pl.getDouble(4); 
	}

	//------------------------------  Build channel entry
	chan_entry ce(name, type, access, dtype, rate);

	//------------------------------  Get source string(s)
	string source;
	for (int i=5; i<nArg; ++i) {
	    source = pl[i];
	    ce.addSource(source, start, end);
	}

	//------------------------------  Put data into temporary
	if (vect_size == vect.size()) {
	    vect_size *= 2;
	    vect.reserve(vect_size);
	}
	vect.push_back(ce);
    }
    cout << "chan_list: " << vect.size() << " channels read from: " 
	 << file << endl;
    merge(vect);
}

//======================================  Reserve space in the channel list
void 
chan_list::reserve(chan_index nMax) {
    mChanList.reserve(nMax);
}
