/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "raw_socket.hh"
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <cstdio>
#include <errno.h>
#include <fcntl.h>
#include <iostream>
#include <iomanip>
#include <sstream>

/*--------------------------------------------------------------------
 *
 *    Select includes depend on posix versions
 *--------------------------------------------------------------------*/
#if _POSIX_C_SOURCE < 200100L
#include <sys/time.h>
#include <sys/types.h>
#else
#include <sys/select.h>
#endif 

using namespace sends;

//======================================  Raw socket constructor
raw_socket::raw_socket(void) {
    mSocket = ::socket(PF_INET, SOCK_STREAM, 0);
    if (mSocket < 0) {
	if (verbose()) perror("Failed to open raw_socket");
	set_error(serr_failure);
    }
}

//======================================  Raw socket constructor
raw_socket::raw_socket(int fd) 
    : mSocket(fd)
{
}

//======================================  Raw socket destructor
raw_socket::~raw_socket(void) {
    close();
}

//======================================  Accept a connection.
raw_socket*
raw_socket::accept(void) {
    raw_socket* rs = 0;

    //----------------------------------  Set up the address return
    sockaddr_in sock;
    sock.sin_family      = AF_INET;
    sock.sin_addr.s_addr = 0;
    sock.sin_port        = 0;
    socklen_t len = 16;

    //----------------------------------  accept the connection
    int fd = ::accept(mSocket, (struct sockaddr*)&sock, &len);
    if (fd < 0) {
	mPeer.clear();
	set_error("accept");
    } else {
	rs = new raw_socket(fd);
	std::ostringstream peer;
	const unsigned char* 
	    p = reinterpret_cast<unsigned char*>(&sock.sin_addr.s_addr);
	peer << int(p[0]) << "." << int(p[1]) << "." << int(p[2]) << "." 
	     << int(p[3]) << ":" << sock.sin_port;
	rs->mPeer = peer.str();
	set_error(serr_OK);
    }
    return rs;
}

//======================================  Bind a socket to a port
raw_socket::sock_err
raw_socket::bind(const std::string& port) {
    sockaddr_in sock;
    sock_err ad_err = parse_addr(port, sock.sin_addr.s_addr, sock.sin_port);
    if (ad_err) {
	return set_error(ad_err, "parse_addr");
    } else if (verbose() > 1) {
	std::cerr << "raw_socket: bind address " << std::hex 
		  << ntohl(sock.sin_addr.s_addr) << std::dec
		  << ":" << ntohs(sock.sin_port) << std::endl;
    }
    sock.sin_family      = AF_INET;
    socklen_t len = 16;
    int rc = ::bind(mSocket, (struct sockaddr*)&sock, len);
    if (rc < 0) return set_error("bind");
    set_error(serr_OK);
    return serr_OK;

}

//======================================  Close the socket
raw_socket::sock_err
raw_socket::close(void) {
    set_error(serr_OK);
    if (mSocket >= 0) {
	::shutdown(mSocket, SHUT_RDWR);
	::close(mSocket);
	mSocket = -1;
    }
    return error();
}

//======================================  Connect to a foreign socket
raw_socket::sock_err
raw_socket::connect(const std::string& peer) {
    mPeer = peer;
    sock_err rc = serr_OK;
    return rc;
}

//======================================  Listen for connection requests
raw_socket::sock_err
raw_socket::listen(const std::string& port, int qlen) {
    sock_err rc = bind(port);
    if (!rc && ::listen(mSocket, qlen)) {
	rc = set_error("listen");
    }
    return rc;
}

//======================================  Read available data
long 
raw_socket::read_available(char* buf, long len) {
    set_error(serr_OK);
    int flags = fcntl (mSocket, F_GETFL, 0);
    fcntl(mSocket, F_SETFL, flags | O_NONBLOCK);
    long nB = ::read(mSocket, buf, len);
    if (nB < 0) {
	set_error("read");
    }
    fcntl(mSocket, F_SETFL, flags);
    return nB;
}

//====================================  Wait routine...
int 
raw_socket::wait_data(double wt) {
    struct timeval tspec, *tptr;
    if (wt < 0) {
	tptr = 0;
    } else {
	long ws = long(wt);
	tspec.tv_sec  = ws;
	tspec.tv_usec = int(100000.0*(wt-double(ws)));
	tptr = &tspec;
    }

    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(mSocket, &readfds);

    set_error(serr_OK);
    int rc = -1;
    while (rc < 0) {
	rc = ::select(mSocket+1, &readfds, 0, 0, tptr);
	if (rc < 0 && errno != EINTR) break;
    }
    if (rc < 0) {
	set_error("select");
    }
    return rc;
}

//======================================  Raw socket constructor
long 
raw_socket::write(const char* buf, long len) {
    long nB = 0;
    while (nB < len) {
	long wrc = ::write(mSocket, buf+nB, len-nB);
	if (wrc > 0) {
	    nB += wrc;
	} else if (set_error("write") != serr_retry) {
	    nB = -1;
	    break;
	}
    }
    if (nB > 0) set_error(serr_OK);
    return nB;
}
